from io import IOBase

from ..api.tool_identity import ToolIdentity
from ..rpc import ReadableByteStream, RPCClient


class ByteStreamEnforcer(IOBase):
    """Ensure that the wrapped file-like object produces bytes

    If it doesn't, then it raises a TypeError with the given error message."""

    def __init__(self, src: ReadableByteStream, err_msg: str):
        self._src = src
        self._err_msg = err_msg

    def read(self, size: int = -1) -> bytes:
        "Read bytes from the wrapped file"
        res = self._src.read(size)

        if not isinstance(res, bytes):
            raise TypeError(self._err_msg)

        return res


class Programmer:
    "A Programmer"

    def __init__(self, rpc: RPCClient, tool_id: str):
        self._rpc = rpc
        self._tool_id = ToolIdentity(tool_id=tool_id)

    def reset_target(self) -> None:
        "Reset the target connected to the device"
        self._rpc.tools_programmer_reset_target(self._tool_id)

    def flash_file(self, src: ReadableByteStream) -> None:
        "Flash the target with the given file"
        enforced = ByteStreamEnforcer(
            src, "flash_file requires a file-like object opened in binary mode"
        )
        self._rpc.tools_programmer_flash(self._tool_id, enforced)



from pydantic import BaseModel

from .tool_request import ToolRequest


class Quantity(BaseModel):
    """
    A physical quantity

    The magnitude is considered to be in SI units (e.g Amps, Volts, Watts) so 1mV
    will be represented with a magnitude of 0.001
    """

    magnitude: float


class QuantityRequest(Quantity, ToolRequest):
    """Model for sending a quantity request to the server"""



from datetime import datetime

from pydantic import BaseModel


class RigLockToken(BaseModel):
    "A token representing holding the lock on a rig"
    token: str


class RigLockHolder(BaseModel):
    "Information about a rig lock holder"
    name: str
    time_of_acquisition: datetime



from pydantic import BaseModel, StrictStr


class WriteParams(BaseModel):
    "The parameters of a write request to a USB Serial tool"
    tool_id: StrictStr
    data: StrictStr


class ReadParams(BaseModel):
    "The parameters of a read request to a USB Serial tool"
    tool_id: StrictStr
    count: int



from enum import Enum
from typing import Optional

from pydantic import BaseModel, validator


class SignalListModel(BaseModel):
    "A list of signals"
    __root__: list[str]


class SignalMode(str, Enum):
    "The mode of a signal"
    INPUT = "input"
    OUTPUT_PP = "output_pp"
    OUTPUT_OD = "output_od"


class SignalState(str, Enum):
    "The state of a signal"
    INACTIVE = "inactive"
    ACTIVE = "active"


class SignalStateModel(BaseModel):
    "Model for the state of a signal"
    __root__: SignalState


class SignalConfig(BaseModel):
    "The configuration of a signal"
    mode: SignalMode
    initial_state: Optional[SignalState] = None


class SignalSetStateParams(BaseModel):
    "Parameters for the tools.gpio.signal.set_state endpoint"
    tool_id: str
    signal: str
    state: SignalState


class SignalConfigureParams(BaseModel):
    "Parameters for the tools.gpio.signal.configure endpoint"
    tool_id: str
    signal: str
    config: SignalConfig


class SignalGetParams(BaseModel):
    "Parameters for the tools.gpio.signal.get_configuration endpoint"
    tool_id: str
    signal: str


class PlaylistState(str, Enum):
    "The state of a GPIO tool's playlist machinery"
    IDLE = "idle"
    PLAYING = "playing"


class PlaylistStateModel(BaseModel):
    "Model for the state of a GPIO tool's playlist machinery"
    __root__: PlaylistState


class PlaylistEntry(BaseModel):
    "An entry in a playlist"
    signal: str
    state: SignalState
    delay_ms: int = 0

    @validator("delay_ms")
    def delay_ms_must_not_be_negative(
        cls, v: int
    ) -> int:  # pylint: disable=no-self-argument # https://github.com/PyCQA/pylint/issues/6900
        """Ensure that delay_ms is positive"""
        if v < 0:
            raise ValueError("delay_ms must not be negative")
        return v


class PlaylistPlayParams(BaseModel):
    "Parameters for the tools.gpio.playlist.play endpoint"
    tool_id: str
    playlist: list[PlaylistEntry]



from dataclasses import dataclass

import polars as pl

from ..api.tool_identity import ToolIdentity
from ..api.tools.logic_analyser import RecordingState
from ..rpc import RPCClient


@dataclass(frozen=True)
class Recording:
    """
    A recording made with the logic analyser tool

    sample_rate_hz: the sampling frequency that was used (Hz)
    data: a polars DataFrame with one column per signal
    """

    sample_rate_hz: int
    data: pl.DataFrame


class LogicAnalyser:
    "A logic analyser tool"

    def __init__(self, rpc: RPCClient, tool_id: str):
        self._rpc = rpc
        self._tool_id = ToolIdentity(tool_id=tool_id)

    def reset(self) -> None:
        """Reset the tool

        Any existing data stored in the tool will be erased,
        and any running recording session will be stopped."""
        self._rpc.rig_tools_reset(self._tool_id)

    def start_recording(
        self, signals: list[str], sample_rate_hz: int, samples: int
    ) -> None:
        """Start a recording

        Request that the logic analyser records the given signals
        at the given sample rate for the given number of samples."""
        self._rpc.tools_logic_analyser_start_recording(
            self._tool_id, signals, sample_rate_hz, samples
        )

    @property
    def data(self) -> Recording:
        "Retrieve the last recording made"
        rec = self._rpc.tools_logic_analyser_get_data(self._tool_id)

        return Recording(
            sample_rate_hz=rec["sample_rate_hz"],
            data=pl.DataFrame(rec["samples"], orient="row", schema=rec["signals"]),
        )

    @property
    def is_recording(self) -> bool:
        "Whether or not the device is currently recording"
        r = self._rpc.tools_logic_analyser_get_recording_state(self._tool_id)
        return r == RecordingState.RECORDING

    @property
    def signals(self) -> tuple[str, ...]:
        "The names of the signals that this tool can record"
        r = self._rpc.tools_logic_analyser_get_signals(self._tool_id)
        return tuple(r)



__all__ = [
    "PlaylistEntry",
    "PlaylistState",
    "SignalConfig",
    "SignalMode",
    "SignalState",
    "GPIOTool",
    "Signal",
    "Playlist",
]

from datetime import timedelta

from ..api.tool_identity import ToolIdentity
from ..api.tools.gpio import (
    PlaylistEntry,
    PlaylistState,
    SignalConfig,
    SignalMode,
    SignalState,
)
from ..rpc import RPCClient


class GPIOTool:
    "A GPIO tool"

    def __init__(self, rpc: RPCClient, tool_id: str):
        self._rpc = rpc
        self._tool_id = ToolIdentity(tool_id=tool_id)

        signal_names = self._rpc.tools_gpio_get_signals(self._tool_id)
        self._signals = {
            name: Signal(name, self._rpc, self._tool_id) for name in signal_names
        }

    def reset(self) -> None:
        """Reset the device

        All GPIO pins will be restored to their default state"""
        self._rpc.rig_tools_reset(self._tool_id)

    @property
    def signals(self) -> dict[str, "Signal"]:
        """A dict of the signals available on this GPIO tool

        The keys of this dictionary are the names of the signals.
        The values are Signal objects."""
        return self._signals

    def play(self, playlist: "Playlist") -> None:
        "Play the given playlist"
        self._rpc.tools_gpio_playlist_play(self._tool_id, playlist._entries)

    @property
    def playlist_running(self) -> bool:
        "Whether a playlist is currently running"
        state = self._rpc.tools_gpio_playlist_get_state(self._tool_id)
        return state == PlaylistState.PLAYING

    def stop_playlist(self) -> None:
        "Stop any currently running playlist"
        self._rpc.tools_gpio_playlist_stop(self._tool_id)


class Signal:
    "A signal on a GPIO tool"

    def __init__(self, name: str, rpc: RPCClient, tool_id: ToolIdentity):
        self._name = name
        self._tool_id = tool_id
        self._rpc = rpc

    @property
    def state(self) -> SignalState:
        """The state of the signal

        See SignalState type for the possible states.  Setting this
        property will modify the state of the signal."""
        return self._rpc.tools_gpio_signal_get_state(self._name, self._tool_id)

    @state.setter
    def state(self, value: SignalState) -> None:
        self._rpc.tools_gpio_signal_set_state(self._name, value, self._tool_id)

    @property
    def config(self) -> SignalConfig:
        """The signal configuration

        See the SignalConfig type for details.  Setting this property
        will modify the signal configuration."""
        return self._rpc.tools_gpio_signal_get_configuration(self._name, self._tool_id)

    @config.setter
    def config(self, config: SignalConfig) -> None:
        self._rpc.tools_gpio_signal_configure(self._name, config, self._tool_id)


class Playlist:
    "A series of timed changes in signal values"

    def __init__(self) -> None:
        self._entries: list[PlaylistEntry] = []

    def set_active(self, signal: str | Signal) -> None:
        """Add an entry to the playlist that makes the given signal active"""
        self._set_state(signal, SignalState.ACTIVE)

    def set_inactive(self, signal: str | Signal) -> None:
        """Add an entry to the playlist that makes the given signal inactive"""
        self._set_state(signal, SignalState.INACTIVE)

    def _set_state(self, signal: str | Signal, state: SignalState) -> None:
        if isinstance(signal, Signal):
            n = signal._name
        else:
            n = signal

        entry = PlaylistEntry(signal=n, state=state)
        self._entries.append(entry)

    def delay(self, delay: timedelta) -> None:
        """Add a time delay to the playlist"""
        try:
            last = self._entries[-1]
        except IndexError:
            raise ValueError("A playlist cannot begin with a delay")

        last.delay_ms += int(delay / timedelta(milliseconds=1))



from enum import Enum

from pydantic import BaseModel


class SignalListModel(BaseModel):
    "List of signals"
    __root__: list[str]


class SampleRatesModel(BaseModel):
    "List of sample rates (Hz)"
    __root__: list[int]


class RecordingState(str, Enum):
    "Logic analyser recording state"
    IDLE = "idle"
    RECORDING = "recording"


class RecordingStateModel(BaseModel):
    "Model for the response from the tools.logic_analyser.get_recording_state endpoint"
    __root__: RecordingState


class StartRecordingParams(BaseModel):
    "Parameters for the tools.logic_analyser.start_recording endpoint"
    tool_id: str
    signals: list[str]
    sample_rate_hz: int
    samples: int


class Recording(BaseModel):
    """A recording made with a logic analyser"""

    # The names of the signals that were sampled
    signals: tuple[str, ...]
    sample_rate_hz: int

    # The samples that were taken in the recording.
    # Indexing: samples[sample_number][signal_index]
    # Where signame_index is the index into the signals
    # tuple from above.
    samples: tuple[tuple[bool, ...], ...]



import os
from dataclasses import dataclass, field, replace
from typing import Any

import requests

from .rpc import RPCClient


class UnauthorizedError(Exception):
    "Invalid API key"


def api_key_from_env() -> str:
    "Load the API key from the environment"
    key = os.getenv("ROCKFACE_API_KEY")
    if key is None:
        raise RuntimeError(
            "Rockface API key not found.  Set the ROCKFACE_API_KEY environment variable."
        )
    return key


def base_url_from_env() -> str:
    "Load the API base URL from the environment"
    return os.getenv("ROCKFACE_API_BASE_URL", default="https://api.rockface.io")


@dataclass
class APIParams:
    "Location and authentication information for the Rockface API"

    api_key: str = field(repr=False, default_factory=api_key_from_env)
    api_base_url: str = field(default_factory=base_url_from_env)


@dataclass(frozen=True)
class RigInfo:
    """Information about a Rig

         id: The ID number of the rig
       name: The human-readable name of the rig
    address: The address of the rig's API
    """

    id: int
    name: str
    address: APIParams

    @classmethod
    def from_dict(cls, src: dict[str, Any], address: APIParams) -> "RigInfo":
        "Build the RigInfo from a JSON-derived dict"

        return cls(id=src["id"], name=src["name"], address=address)

    def build_rpc(self) -> RPCClient:
        "Build an RPCClient for interacting with this rig"
        return RPCClient(
            base_url=self.address.api_base_url, api_key=self.address.api_key
        )


@dataclass
class RockfaceClient:
    "A client for accessing the Rockface API"
    params: APIParams

    @property
    def rigs(self) -> list[RigInfo]:
        "List of the available rigs"

        resp = requests.get(
            f"{self.params.api_base_url}/rigs",
            headers={"rockface-api-token": self.params.api_key},
        )

        if resp.status_code == requests.codes.unauthorized:
            raise UnauthorizedError()

        resp.raise_for_status()

        result = []
        for rig in resp.json():
            address = replace(
                self.params, api_base_url=f"{self.params.api_base_url}/rigs/{rig['id']}"
            )

            result.append(RigInfo.from_dict(rig, address))

        return result



from rockface.api.quantity import Quantity
from rockface.api.tools.psu import CurrentLimitMode, PSUMode, PSUSpec, PSUState

from ..api.tool_identity import ToolIdentity
from ..rpc import RPCClient


class PSUConfig:
    "The configuration of a power supply tool"

    def __init__(self, psu: "PSUTool"):
        self._psu = psu

    @property
    def voltage(self) -> float:
        """The voltage setpoint

        This is the voltage that the power supply will attempt to
        drive its output to.  Set this property to modify the
        setpoint."""
        return self._psu._rpc.tools_psu_get_voltage_setpoint(
            self._psu._tool_id
        ).magnitude

    @voltage.setter
    def voltage(self, voltage: float) -> None:
        self._psu._rpc.tools_psu_set_voltage_setpoint(
            self._psu._tool_id, Quantity(magnitude=voltage)
        )

    @property
    def current_limit(self) -> float:
        """The current limit magnitude

        This is the maximum current that the power supply will
        deliver to the load.  Set this property to modify it."""
        return self._psu._rpc.tools_psu_get_current_limit(self._psu._tool_id).magnitude

    @current_limit.setter
    def current_limit(self, current_limit: float) -> None:
        self._psu._rpc.tools_psu_set_current_limit(
            self._psu._tool_id, Quantity(magnitude=current_limit)
        )

    @property
    def current_limit_mode(self) -> CurrentLimitMode:
        """The type of current limiting in use

        See the CurrentLimitMode type for a description of possible
        values.   Set this property to change mode.
        """
        return self._psu._rpc.tools_psu_get_current_limit_mode(self._psu._tool_id)

    @current_limit_mode.setter
    def current_limit_mode(self, current_limit_mode: CurrentLimitMode) -> None:
        self._psu._rpc.tools_psu_set_current_limit_mode(
            self._psu._tool_id, current_limit_mode
        )


class PSUTool:
    "A PSU tool"

    def __init__(self, rpc: RPCClient, tool_id: str):
        self._rpc = rpc
        self._tool_id = ToolIdentity(tool_id=tool_id)

    def reset(self) -> None:
        """Reset the tool

        The existing configuration will be lost and the output will be disabled
        """
        self._rpc.rig_tools_reset(self._tool_id)

    @property
    def _state(self) -> PSUState:
        return self._rpc.tools_psu_get_state(self._tool_id)

    @property
    def enabled(self) -> bool:
        """Whether or not the output is enabled"""
        return self._state.enabled

    @enabled.setter
    def enabled(self, enable: bool) -> None:
        if enable:
            self._rpc.tools_psu_enable(self._tool_id)
        else:
            self._rpc.tools_psu_disable(self._tool_id)

    @property
    def tripped(self) -> bool:
        """Whether the tool has disabled its output as a result of exceeding its current limit"""
        return self._state.tripped

    @tripped.setter
    def tripped(self, tripped: bool) -> None:
        if tripped:
            raise ValueError("tripped state can only be cleared")
        self._rpc.tools_psu_clear_trip(self._tool_id)

    @property
    def current(self) -> float:
        """The output current measured by the power supply, in amps

        For the configured current limit see PSUTool.config
        """
        return self._state.current

    @property
    def voltage(self) -> float:
        """The output voltage measured by the power supply, in volts

        For the configured voltage see PSUTool.config
        """
        return self._state.voltage

    @property
    def in_constant_current(self) -> bool:
        """Whether or not the device is in constant current mode"""
        return self._state.mode == PSUMode.CONSTANT_CURRENT

    @property
    def in_constant_voltage(self) -> bool:
        """Whether or not the device is in constant voltage mode"""
        return self._state.mode == PSUMode.CONSTANT_VOLTAGE

    @property
    def spec(self) -> PSUSpec:
        """The tool specification"""
        return self._rpc.tools_psu_get_capabilities(self._tool_id)

    @property
    def config(self) -> PSUConfig:
        """The configuration of the tool"""
        return PSUConfig(self)



import json
import os

FlagSet = frozenset[str]


_SUPPORTED_FLAGS = FlagSet()

_DEFAULT_FLAGS = FlagSet()


class InvalidFeatureFlag(Exception):
    "An invalid feature name flag was provided"


def _load_feature_flags() -> set[str]:
    flags = json.loads(os.getenv("ROCKFACE_FEATURE_FLAGS", default="[]"))

    if not isinstance(flags, list):
        raise ValueError("Malformed feature flag environment variable")

    for flag in flags:
        if flag not in _SUPPORTED_FLAGS:
            raise InvalidFeatureFlag(
                f"Unknown feature flag in environment variable: '{flag}'"
            )

    return set(flags) | _DEFAULT_FLAGS


FEATURE_FLAGS = _load_feature_flags()


def flag_enabled(flag_name: str) -> bool:
    "Determine if the named feature flag is enabled"

    if flag_name not in _SUPPORTED_FLAGS:
        raise InvalidFeatureFlag(f"{flag_name} is not a known flag")

    return flag_name in FEATURE_FLAGS



import base64
import time
from typing import Any, Protocol

import requests
from pydantic import parse_obj_as

from .api import rig as api_rig
from .api import rig_lock as api_rig_lock
from .api.quantity import Quantity, QuantityRequest
from .api.tool_identity import ToolIdentity
from .api.tools import gpio as api_gpio
from .api.tools import logic_analyser as api_la
from .api.tools import psu as api_psu
from .api.tools import usb_serial as api_usb_serial


class ReadableByteStream(Protocol):
    "Protocol for a file-like object that produces bytes when read"

    def read(self, size: int) -> bytes:
        pass


MultiPartFile = tuple[str, ReadableByteStream, str]


class RPCClient:
    "Adapter for interacting with the rockface HTTP API"

    def __init__(self, base_url: str, api_key: str, rig_lock_token: str | None = None):
        self._base_url = base_url
        self._api_key = api_key
        self._last_request_time: None | float = None
        self._session = requests.Session()
        self._rig_lock_token = rig_lock_token

    def clone_with_rig_lock_token(self, token: str) -> "RPCClient":
        "Create a new RPCClient that uses the given rig lock token"
        return RPCClient(
            base_url=self._base_url, api_key=self._api_key, rig_lock_token=token
        )

    def call(
        self,
        method: str,
        params: Any = None,
        files: None | list[tuple[str, MultiPartFile]] = None,
    ) -> Any:
        """Call an RPC method

        method: Name of the endpoint
        params: Parameters for the call.
                Can be a pydantic type, or just a dictionary.
        files: List of files to provide with the request.
        """
        kw: dict[str, Any] = {"files": [] if files is None else files}

        if hasattr(params, "dict"):
            params = params.dict()

        if params is not None:
            if files is not None:
                """Cannot send JSON request simultaneously with files,
                so send the parameters as form data instead."""
                kw["data"] = params
            else:
                kw["json"] = params

        kw["headers"] = {"rockface-api-token": self._api_key}

        if self._rig_lock_token is not None:
            kw["headers"]["rockface-rig-lock-token"] = self._rig_lock_token

        if (
            self._last_request_time is not None
            and self._last_request_time - time.monotonic() > 60
        ):
            """The server's keepalive timeout is set to 120s, and we have not talked
            to it in nearly that time.  Unfortunately requests has an issue if
            we an issue a request at roughly the same time as the connection timeout
            and restart the session if we have potentially timed-out.
            https://github.com/psf/requests/issues/4664"""
            self._session.close()
            self._session = requests.Session()

        r = self._session.post(f"{self._base_url}/{method}", **kw)
        self._last_request_time = time.monotonic()
        r.raise_for_status()
        return r.json()

    def rig_tools_list(self) -> list[api_rig.ToolIdentity]:
        "List the tools on the rig"
        resp = self.call("rig.tools.list")
        return parse_obj_as(list[api_rig.ToolIdentity], resp)

    def rig_tools_reset(self, tool_id: ToolIdentity) -> None:
        self.call("rig.tools.reset", tool_id)

    def rig_lock_acquire(self, owner: str) -> str:
        "Acquire the rig lock"
        resp = self.call("rig.lock.acquire", {"owner": owner})
        return api_rig_lock.RigLockToken.parse_obj(resp).token

    def rig_lock_release(self) -> None:
        "Release the rig lock"
        self.call("rig.lock.release")

    def rig_lock_force_release(self, holder_name: str | None = None) -> None:
        "Force the release of lock, even if we are not the current holder"
        params = {}
        if holder_name is not None:
            params["holder_name"] = holder_name

        self.call("rig.lock.force_release", params)

    def rig_lock_get_holder(self) -> api_rig_lock.RigLockHolder:
        "Get the current holder of the rig lock"
        resp = self.call("rig.lock.get_holder")
        return api_rig_lock.RigLockHolder.parse_obj(resp)

    def tools_gpio_identify(self) -> ToolIdentity:
        "Retrieve the identity of a GPIO tool"
        resp = self.call("tools.gpio.identify")
        return ToolIdentity.parse_obj(resp)

    def tools_gpio_get_signals(self, tool_id: ToolIdentity) -> list[str]:
        "Get the signals available on a GPIO tool"
        resp = self.call("tools.gpio.get_signals", tool_id)
        return api_gpio.SignalListModel.parse_obj(resp).__root__

    def tools_gpio_signal_configure(
        self, signal: str, config: api_gpio.SignalConfig, tool_id: ToolIdentity
    ) -> None:
        "Configure a signal on a GPIO tool"
        param = api_gpio.SignalConfigureParams(
            tool_id=tool_id.tool_id, signal=signal, config=config
        )
        self.call("tools.gpio.signal.configure", param)

    def tools_gpio_signal_get_configuration(
        self, signal: str, tool_id: ToolIdentity
    ) -> api_gpio.SignalConfig:
        "Get the configuration of a signal on a GPIO tool"
        param = api_gpio.SignalGetParams(tool_id=tool_id.tool_id, signal=signal)
        resp = self.call("tools.gpio.signal.get_configuration", param)
        return api_gpio.SignalConfig.parse_obj(resp)

    def tools_gpio_signal_set_state(
        self, signal: str, state: api_gpio.SignalState, tool_id: ToolIdentity
    ) -> None:
        "Set the state of a signal on a GPIO tool"
        param = api_gpio.SignalSetStateParams(
            tool_id=tool_id.tool_id, signal=signal, state=state
        )
        self.call("tools.gpio.signal.set_state", param)

    def tools_gpio_signal_get_state(
        self, signal: str, tool_id: ToolIdentity
    ) -> api_gpio.SignalState:
        "Get the state of a signal on a GPIO tool"
        param = api_gpio.SignalGetParams(tool_id=tool_id.tool_id, signal=signal)
        resp = self.call("tools.gpio.signal.get_state", param)
        return api_gpio.SignalStateModel.parse_obj(resp).__root__

    def tools_gpio_playlist_play(
        self, tool_id: ToolIdentity, playlist: list[api_gpio.PlaylistEntry]
    ) -> None:
        "Play a playlist on a GPIO tool"
        param = api_gpio.PlaylistPlayParams(tool_id=tool_id.tool_id, playlist=playlist)
        self.call("tools.gpio.playlist.play", param)

    def tools_gpio_playlist_get_state(
        self, tool_id: ToolIdentity
    ) -> api_gpio.PlaylistState:
        "Get the playlist playback state of a GPIO tool"
        resp = self.call("tools.gpio.playlist.get_state", tool_id)
        return api_gpio.PlaylistStateModel.parse_obj(resp).__root__

    def tools_gpio_playlist_stop(self, tool_id: ToolIdentity) -> None:
        "Make a GPIO tool stop playing its playlist"
        self.call("tools.gpio.playlist.stop", tool_id)

    def tools_programmer_reset_target(self, tool_id: ToolIdentity) -> None:
        "Reset the target connected to a programmer tool"
        self.call("tools.programmer.reset_target", tool_id)

    def tools_programmer_flash(
        self, tool_id: ToolIdentity, src: ReadableByteStream
    ) -> None:
        "Load the given hex file into the target connected to a programmer tool"
        self.call(
            "tools.programmer.flash",
            tool_id,
            files=[("file", ("src.hex", src, "application/octet-stream"))],
        )

    def tools_logic_analyser_get_signals(self, tool_id: ToolIdentity) -> list[str]:
        "Get the list of signals on a logic analyser tool"
        r = self.call("tools.logic_analyser.get_signals", tool_id)
        return api_la.SignalListModel.parse_obj(r).__root__

    def tools_logic_analyser_get_sample_rates(self, tool_id: ToolIdentity) -> list[int]:
        "Get the sample rates a logic analyser tool supports"
        r = self.call("tools.logic_analyser.get_sample_rates", tool_id)
        return api_la.SampleRatesModel.parse_obj(r).__root__

    def tools_logic_analyser_get_recording_state(
        self, tool_id: ToolIdentity
    ) -> api_la.RecordingState:
        "Get the recording state of a logic analyser"
        r = self.call("tools.logic_analyser.get_recording_state", tool_id)
        return api_la.RecordingStateModel.parse_obj(r).__root__

    def tools_logic_analyser_start_recording(
        self,
        tool_id: ToolIdentity,
        signals: list[str],
        sample_rate_hz: int,
        samples: int,
    ) -> None:
        "Make a logic analyser tool start recording"
        param = api_la.StartRecordingParams(
            tool_id=tool_id.tool_id,
            signals=signals,
            sample_rate_hz=sample_rate_hz,
            samples=samples,
        )
        self.call("tools.logic_analyser.start_recording", param)

    def tools_logic_analyser_get_data(self, tool_id: ToolIdentity) -> Any:
        "Retrieve the recorded data from a logic analyser tool"
        return self.call("tools.logic_analyser.get_data", tool_id)

    def tools_psu_enable(self, tool_id: ToolIdentity) -> None:
        "Enable the output of a power supply tool"
        self.call("tools.psu.enable", tool_id)

    def tools_psu_disable(self, tool_id: ToolIdentity) -> None:
        "Disable the output of a power supply tool"
        self.call("tools.psu.disable", tool_id)

    def tools_psu_get_state(self, tool_id: ToolIdentity) -> api_psu.PSUState:
        "Get the state of a power supply tool"
        r = self.call("tools.psu.get_state", tool_id)
        return api_psu.PSUState.parse_obj(r)

    def tools_psu_get_capabilities(self, tool_id: ToolIdentity) -> api_psu.PSUSpec:
        "Get the capabilities of a power supply tool"
        r = self.call("tools.psu.get_capabilities", tool_id)
        return api_psu.PSUSpec.parse_obj(r)

    def tools_psu_set_voltage_setpoint(
        self,
        tool_id: ToolIdentity,
        quantity: Quantity,
    ) -> None:
        "Set the voltage setpoint of a power supply tool"
        self.call(
            "tools.psu.set_voltage_setpoint",
            QuantityRequest(
                tool_id=tool_id.tool_id,
                magnitude=quantity.magnitude,
            ),
        )

    def tools_psu_get_voltage_setpoint(self, tool_id: ToolIdentity) -> Quantity:
        "Get the voltage setpoint of a power supply tool"
        return Quantity.parse_obj(self.call("tools.psu.get_voltage_setpoint", tool_id))

    def tools_psu_get_current_limit(self, tool_id: ToolIdentity) -> Quantity:
        "Get the current limit of a power supply tool"
        return Quantity.parse_obj(self.call("tools.psu.get_current_limit", tool_id))

    def tools_psu_set_current_limit(
        self, tool_id: ToolIdentity, quantity: Quantity
    ) -> None:
        "Set the current limit of a power supply tool"
        self.call(
            "tools.psu.set_current_limit",
            QuantityRequest(
                tool_id=tool_id.tool_id,
                magnitude=quantity.magnitude,
            ),
        )

    def tools_psu_clear_trip(self, tool_id: ToolIdentity) -> None:
        self.call("tools.psu.clear_trip", tool_id)

    def tools_psu_get_current_limit_mode(
        self, tool_id: ToolIdentity
    ) -> api_psu.CurrentLimitMode:
        return api_psu.CurrentLimitMode(
            self.call("tools.psu.get_current_limit_mode", tool_id)["current_limit_mode"]
        )

    def tools_psu_set_current_limit_mode(
        self, tool_id: ToolIdentity, current_limit_mode: api_psu.CurrentLimitMode
    ) -> None:
        self.call(
            "tools.psu.set_current_limit_mode",
            api_psu.CurrentLimitModeParams(
                tool_id=tool_id.tool_id, current_limit_mode=current_limit_mode
            ),
        )

    def tools_usb_serial_is_available(self, tool_id: ToolIdentity) -> bool:
        "Determine if the USB device associated with a USB serial tool is available"
        return bool(self.call("tools.usb_serial.is_available", tool_id))

    def tools_usb_serial_write(self, tool_id: ToolIdentity, data: bytes) -> None:
        "Write some data via a USB serial tool"
        ascii_encoded = base64.b64encode(data).decode("ascii")
        param = api_usb_serial.WriteParams(tool_id=tool_id.tool_id, data=ascii_encoded)
        self.call("tools.serial.write", param)

    def tools_usb_serial_read(self, tool_id: ToolIdentity, count: int) -> bytes:
        "Read some data from a USB serial tool"
        param = api_usb_serial.ReadParams(tool_id=tool_id.tool_id, count=count)
        return base64.b64decode(self.call("tools.serial.read", param))

    def tools_usb_serial_get_descriptors(self, tool_id: ToolIdentity) -> dict[str, str]:
        "Get the USB descriptors associated with a USB serial tool"
        return dict(self.call("tools.usb_serial.get_descriptors", tool_id))



from ..api.tool_identity import ToolIdentity
from ..rpc import RPCClient


class USBSerial:
    """A USB Serial Tool"""

    def __init__(self, rpc: RPCClient, tool_id: str):
        self._tool_id = ToolIdentity(tool_id=tool_id)
        self._rpc = rpc

    @property
    def is_available(self) -> bool:
        """Whether the usb serial device is available for read/write operations"""
        return self._rpc.tools_usb_serial_is_available(self._tool_id)

    def write(self, data: bytes) -> None:
        """Write the given bytes to the USB serial device"""
        self._rpc.tools_usb_serial_write(self._tool_id, data)

    def read(self, count: int) -> bytes:
        """Read count bytes from the USB serial device"""
        return self._rpc.tools_usb_serial_read(self._tool_id, count)

    @property
    def descriptors(self) -> dict[str, str]:
        """The usb device descriptors"""
        return self._rpc.tools_usb_serial_get_descriptors(self._tool_id)



from enum import Enum
from math import isclose

from pydantic import BaseModel, confloat, root_validator

from ...api.tool_request import ToolRequest


class CurrentLimitMode(str, Enum):
    """The behaviour of a PSU tool when the current limit is reached

    LIMIT: The power supply will enter into "constant current" mode
           if the limit is reached -- i.e. it will reduce the voltage
           it is delivering to the load to keep the current below the
           limit.
    TRIP: If the current delivered to the load exceeds the limit, then
          the power supply will turn off."""

    LIMIT = "limit"
    TRIP = "trip"


class CurrentLimitModeParams(ToolRequest):
    "Parameters for the tools.psu.set_current_limit_mode endpoint"
    current_limit_mode: CurrentLimitMode


class PSUMode(str, Enum):
    """The operating mode of a PSU tool."""

    CONSTANT_CURRENT = "constant current"
    CONSTANT_VOLTAGE = "constant voltage"


class PSUState(BaseModel):
    """
    The state of a PSU tool

    Attributes:
        voltage (float): The output voltage in volts.
        current (float): The output current in amps.
        enabled (bool): True if the output of the PSU is enabled.
        tripped (bool): True if the PSU has tripped.
        mode (PSUMode): Operating mode of the tool.
    """

    voltage: float
    current: float
    enabled: bool
    tripped: bool
    mode: PSUMode


class Range(BaseModel):
    "An inclusive range."
    min: float
    max: float

    @root_validator
    def check_min_not_gt_max(cls, values: dict[str, float]) -> dict[str, float]:
        """Ensure that min <= max"""
        # pylint: disable=no-self-argument,no-self-use
        #   Disable pylint as it is unable to determine that root_validators are classmethods,
        #   see: https://github.com/samuelcolvin/pydantic/issues/568)
        _min, _max = values.get("min"), values.get("max")
        assert _min is not None and _max is not None

        assert not _min > _max
        return values

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Range):
            raise NotImplementedError()
        return isclose(self.min, other.min) and isclose(self.max, other.max)


class CurrentCapabilities(Range):
    "The current capabilities of a power supply, all fields reported in Amps"
    limit_resolution: float
    min: confloat(le=0)  # type: ignore # mypy considers these invalid type annotations
    max: confloat(ge=0)  # type: ignore # mypy considers these invalid type annotations


class VoltageCapabilities(Range):
    "The voltage capabilities of a power supply, all fields reported in Volts"
    setpoint_resolution: confloat(gt=0)  # type: ignore # mypy considers these invalid type annotations
    sense_resolution: confloat(gt=0)  # type: ignore # mypy considers these invalid type annotations

    @root_validator
    def check_min_not_gt_max(cls, values: dict[str, float]) -> dict[str, float]:
        """Ensure that min <= max"""
        # pylint: disable=no-self-argument,no-self-use
        #   Disable pylint as it is unable to determine that root_validators are classmethods,
        #   see: https://github.com/samuelcolvin/pydantic/issues/568)
        _min, _max = values.get("min"), values.get("max")
        assert _min is not None and _max is not None

        assert not _min > _max
        return values


class PSUSpec(BaseModel):
    "The specifications of a power supply"
    voltage: VoltageCapabilities
    current: CurrentCapabilities



import functools
import os
import socket
import sys
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Callable, TypeVar

from .client import APIParams, RigInfo, RockfaceClient
from .rpc import RPCClient
from .tools.gpio import GPIOTool
from .tools.logic_analyser import LogicAnalyser
from .tools.programmer import Programmer
from .tools.psu import PSUTool
from .tools.usb_serial import USBSerial


@dataclass(frozen=True)
class RigSelector:
    "A description of a method of selecting a rig"
    name: str | None = None

    def matches(self, rig_info: RigInfo) -> bool:
        return self.name == rig_info.name


tool_factories = {
    "gpio": GPIOTool,
    "programmer": Programmer,
    "logic_analyser": LogicAnalyser,
    "psu": PSUTool,
    "serial": USBSerial,
}

T = TypeVar("T")


def cache(func: Callable[..., T]) -> T:
    "Workaround for mypy's cache problems"
    # https://github.com/python/mypy/issues/5858
    return functools.lru_cache()(func)  # type: ignore


class BoundRig:
    """A rockface rig

    This class is not intended to be instantiated directly.  Use the
    ``Rig`` class to get an instance of this class.
    """

    def __init__(
        self,
        rpc: RPCClient,
        info: RigInfo,
        tool_factories: dict[str, Any] = tool_factories,
    ):
        self._tool_factories = tool_factories
        self._rpc: RPCClient | None = rpc
        self._info = info

    @classmethod
    def from_unlocked_rpc(cls, unlocked_rpc: RPCClient, info: RigInfo) -> "BoundRig":
        owner = f"{socket.gethostname()}-{os.getpid()}"
        token = unlocked_rpc.rig_lock_acquire(owner)
        rpc = unlocked_rpc.clone_with_rig_lock_token(token)
        return cls(rpc, info)

    @classmethod
    def from_selector(
        cls,
        selector: RigSelector,
        rockface_client: RockfaceClient,
    ) -> "BoundRig":

        matching_rigs = (rig for rig in rockface_client.rigs if selector.matches(rig))
        try:
            rig = next(matching_rigs)
        except StopIteration:
            raise LookupError(f"No rig found matching {selector}")
        return cls.from_unlocked_rpc(rig.build_rpc(), rig)

    def __enter__(self) -> "BoundRig":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.release()

    def release(self) -> None:
        "Release the lock on this rig"
        if self._rpc is None:
            raise Exception("Rig lock is already released")

        self._rpc.rig_lock_release()
        self._rpc = None

    def __del__(self) -> None:
        if hasattr(self, "_rpc") and self._rpc is not None:
            self.release()

    @property
    def info(self) -> RigInfo:
        return self._info

    @property
    def name(self) -> str:
        "The name of the rig"
        assert self._info is not None
        return self._info.name

    @property
    @cache
    def tools(self) -> dict[str, Any]:
        "The set of tools we have on the rig"
        assert self._rpc is not None
        resp = self._rpc.rig_tools_list()
        tools = {}

        for desc in resp:
            if desc.kind not in self._tool_factories:
                print(  # noqa: PrintFound
                    f"Warning: tool type '{desc.kind}' is ignored by this version of the rockface library.",
                    file=sys.stderr,
                )
                continue
            tool = self._tool_factories[desc.kind](self._rpc, desc.id)
            tools[desc.id] = tool

        return tools


class Rig:
    """'Construction' class for a rig

    To construct, use the ``find_by_name`` classmethod.

    This method will find your rig, acquire a lock on it, and
    return the rig to you to use.

    Use the returned Rig as a context manager to scope the lock
    within your code:

      with Rig.find_by_name("parrot") as rig:
          # Use ``rig`` to do whatever you like
          ...

    If you don't want to use this as a context manager, you don't
    have to, but the lock will only be dropped when the returned
    rig object is garbage collected, which can result in surprising
    situations.

    Taking the lock can be delegated to a later point in time
    by providing the ``lock=False`` parameter.  Use the resulting
    object as a context manager to lock the rig:

      rig = Rig.find_by_name("cheese")
      with rig:
          # `rig` is locked within this context
    """

    def __init__(self, *args: Any, **kw: Any) -> None:
        self._args = args
        self._kw = kw
        self._bound: BoundRig | None = None

    def __enter__(self) -> "BoundRig":
        "Select and lock a rig"
        if self._bound is not None:
            raise Exception("Rig is already bound")
        self._bound = BoundRig.from_selector(*self._args, **self._kw)
        return self._bound

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        "Release the locked rig"
        if self._bound is not None:
            self._bound.release()
            self._bound = None

    @classmethod
    def _by_selector(
        cls,
        selector: RigSelector,
        lock: bool = True,
        bound_rig_factory: Callable[..., BoundRig] = BoundRig.from_selector,
        rockface_client_factory: Callable[..., RockfaceClient] = RockfaceClient,
        api_key: str | None = None,
        api_base_url: str | None = None,
    ) -> "Rig | BoundRig":
        "Use the given selector to find the rig"
        kw = {"api_key": api_key, "api_base_url": api_base_url}
        api_params = APIParams(**{k: v for k, v in kw.items() if v is not None})
        client = rockface_client_factory(api_params)
        if lock:
            "Lock the rig immediately"
            return bound_rig_factory(selector, rockface_client=client)
        return cls(selector, rockface_client=client)

    @classmethod
    def find_by_name(cls, name: str, lock: bool = True, **kw: Any) -> "Rig | BoundRig":
        "Use the rig with the given name"
        selector = RigSelector(name=name)
        return cls._by_selector(selector, lock=lock, **kw)
